;
; Spring Piano V2.0
; (c) 2009 by Nilquader of SPRING
; Z80 Assembler Code for the Amstrad NC100
;
; Missing: octave stop at no. 9
;
	call txtclearwindow
	ld hl, welcmsg
	call print
	call refreshdisp
.loop ; main loop
	call readkeys
	
	ld b,a ; save A
	and &40 ; 6th bit set? Command key!
	ld a,b
	jr z, loop1
	call cmdkey
	jr loop
.loop1
	; check first key
	push de ; save d/e (tones)
	ld a, d
	ld hl,lastkey1 ; did key change?
	cp (hl)
	jr z, loop2 ; key did not change, check second key
	ld (hl),a
	
	ld de, &FFFF
	cp &80
	call nz,lookuptone ; look up period value
	
	ld c, &50 ; program sound register (1)
	out (c), e
	inc c
	out (c), d

.loop2
	; check secound key
	pop de ; restore d/e (tones)
	ld hl,polyswitch
	ld a,(hl)
	or a
	jr z,loop ; omit second channel if polyphony is switched off
	
	ld a,e
	ld hl, lastkey2
	cp (hl)
	jr z, loop ; key did not change, restart main loop
	ld (hl),a
	
	ld de, &FFFF
	cp &80
	call nz,lookuptone ; look up period value
	
	ld c, &52 ; program sound register (2)
	out (c), e
	inc c
	out (c), d

	jr loop
	; end.


; read keyboard and translate to piano keys
; OUT: a (command key or &80 if none), de (up to 2 note keys)
; destroys BC
.readkeys
	ld de, &8080
	ld hl, keytable
	ld c,&b0
.readkeys1 ; loop repeats for each port
	in a, (c)
	ld b, 8
.readkeys2 ; loop repeats for each bit
	rra
	jr c, readkeys3 ; save keycode when key pressed, hl is pointer to place in k-table
.readkeys2a
	inc hl
	dec b
	jr nz, readkeys2
	inc c
	ld a,c
	cp &ba
	ld a, &80 ; indicate: no command key OR no suitable key found
	ret z
	jr readkeys1
.readkeys3
	; 3 possible actions:
	; (1) Command key -> return in a and quit
	; (2) first note key -> save in d and continue
	; (3) second note key -> save in e and quit
	db &dd, &6f ; ld ixl, a ; save a 
	ld a,(hl) ; get keycode
	bit 6,a ; check for command key
	ret nz ; return if key is command key (CASE 1)
	bit 7,d ; check if d register is "empty"(=&80)
	jr z,readkeys4 ; already set, check next
	ld d,a ; save key in d...
	db &dd, &7d ; ld a, ixl ; restore a
	jr readkeys2a ; ...and continue (CASE 2)
.readkeys4
	ld e,a ; save key in e
	ld a, &80 ; indicate: no command key...
	ret ; ...and return (CASE 3)

; look up period values
; IN a (tone number), OUT de (period values for sound register)
; DESTROYS hl
.lookuptone
	; calculate tone number according to octave
	bit 5,a ; check if tone is from lower manual
	ld hl,octaveup
	jr z,lookuptone2
	ld hl,octavelo
	and &1F ; clear bit 5
.lookuptone2
	ld d, (hl)
	add a,d ; multiply by 12
	add a,d
	add a,d
	add a,d
	add a,d
	add a,d
	add a,d
	add a,d
	add a,d
	add a,d
	add a,d
	add a,d
	; look up tone position in table
	ld hl, freqtbl
	add a,a ; A = A * 2
	
	add a,l ; add A to HL
	ld l,a
	jr nc, lookuptone1 ; no overflow
	inc h ; carry
.lookuptone1
	ld e, (hl)
	inc hl
	ld d, (hl)
	ret
	
; command keys
.cmdkey
	;ld a,b ; restore A
	cp &41 ; ESC -> Quit program
	jr nz, cmdkey1
	pop hl ; take return address from stack and throw away
	ret ; return to OS (aka QUIT)
.cmdkey1
	cp &42 ; DEL> -> Octave down (upper manual)
	jr nz, cmdkey2
	ld hl, octaveup
	dec (hl)
	push af
	call refreshdisp
.cmdkey1a
	in a,(&B6) 
	and &04 ; 2nd bit of Port &B6
	jr nz,cmdkey1a ; wait for key release
	pop af
	ret
.cmdkey2
	cp &43 ; DEL< -> Octave up (upper manual)
	jr nz, cmdkey3
	ld hl, octaveup
	inc (hl)
	push af
	call refreshdisp
.cmdkey2a
	in a,(&B9) 
	and &04 ; 2nd bit of Port &B9
	jr nz,cmdkey2a ; wait for key release	
	pop af
	ret
.cmdkey3
	cp &44 ; Cursor left -> Octave down (lower manual)
	jr nz, cmdkey4
	ld hl, octavelo
	dec (hl)
	push af
	call refreshdisp
.cmdkey3a
	in a,(&B0) 
	and &08 ; 3rd bit of Port &B0
	jr nz,cmdkey3a ; wait for key release
	pop af
	ret		
.cmdkey4
	cp &45 ; Cursor right -> Octave up (lower manual)
	jr nz, cmdkey5
	ld hl, octavelo
	inc (hl)
	push af
	call refreshdisp
.cmdkey4a
	in a,(&B6) 
	and &08 ; 3rd bit of Port &B6
	jr nz,cmdkey4a ; wait for key release	
	pop af
	ret
.cmdkey5
	cp &46 ; MENU -> Polyphony switch
	jr nz, cmdkey6
	ld hl, polyswitch
	ld a, (hl)
	xor 1 ; invert 1st bit
	ld (hl),a
	call refreshdisp
.cmdkey5a
	in a,(&B7)
	and &10 ; 4th bit of Port &B7
	jr nz, cmdkey5a ; wait for key release	
.cmdkey6
	ret
	
; Refresh display
.refreshdisp
	; print octave numbers
	ld hl,&4C00 ; Line &00, Column &4C
	call txtsetcursor
	ld hl, octaveup
	ld a,(hl)
	cp &ff
	jr nz, refreshdisp1
	xor a
	ld (hl),a
.refreshdisp1
	call digit
	ld hl,&4C01 ; Line &01, Column &4C
	call txtsetcursor
	ld hl, octavelo
	ld a,(hl)
	cp &ff
	jr nz, refreshdisp2
	xor a
	ld (hl),a
.refreshdisp2
	call digit
	; print polyphony switch
	ld hl, &4C03 ; Line &03, Column &4C
	call txtsetcursor
	
	ld hl, polyswitch
	ld a, (hl)
	ld hl, txtoff
	or a
	jr z, refreshdisp3
	ld hl,txton
.refreshdisp3
	call print
	ret

; print single digit
.digit
	add a,"0"
	jp charout

;
; variables
;
.lastkey1
	db &80
.lastkey2
	db &80
.octaveup
	db &04
.octavelo
	db &04
.polyswitch
	db &01
;
; constants / tables
;
.welcmsg ; welcome message
	defb "  2 3   5 6 7   9 0   =                                       UPPER Octave:",10,13
	defb " Q W E R T Y U I O P [ ]                                      LOWER Octave:",10,13
	defb 10,13
	defb " UPPER/LOWER MANUALS          ",12,"SPRING PIANO V2.1",12,"                  Polyphony:",10,13
	defb 10,13
	defb "  S D   G H J   L ;   #",10,13
	defb " Z X C V B N M , . / ^ ^",10,13
	defb "Press DELETE & CURSOR to change octave, MENU to toggle polyphony and ESC to Quit",0 
	
.txton
	defb "On ",0
.txtoff
	defb "Off",0

.keytable ; keyboard to piano key (note number) translation table
; Bit:  0   1   2   3   4   5   6   7
	db &80,&31,&80,&44,&80,&80,&80,&80 ; Port &B0
	db &80,&80,&41,&80,&80,&80,&06,&80 ; Port &B1
	db &80,&80,&80,&80,&80,&80,&80,&80 ; Port &B2
	db &03,&01,&00,&02,&04,&80,&21,&23 ; Port &B3
	db &80,&80,&20,&22,&80,&80,&05,&80 ; Port &B4
	db &80,&80,&27,&25,&07,&09,&26,&24 ; Port &B5
	db &08,&80,&42,&45,&32,&30,&28,&29 ; Port &B6
	db &12,&0A,&80,&33,&46,&0B,&2B,&80 ; Port &B7
	db &80,&80,&13,&11,&80,&0C,&2A,&2C ; Port &B8
	db &0F,&0D,&43,&10,&2F,&2D,&0E,&2E ; Port &B9
;
; Bit 0-4 : Note value
; Bit   5 : upper(0)/lower(1) manual
; Bit   6 : note(0)/command(1) key
; Bit   7 : unimportant key
; 
; Command key reference:
; &41 : ESC  -> Quit program
; &42 : DEL> -> Octave down (upper manual)
; &43 : DEL< -> Octave up (upper manual)
; &44 : <=   -> Octave down (lower manual)
; &45 : =>   -> Octave up (lower manual)
; &46 : MENU -> Polyphony switch
;

.freqtbl ; frequency table (in external file, precalculated)
	include "freqtbl.asm"

include "nc100io.asm"
